"""Compensation constants."""
DOMAIN = "compensation"

SENSOR = "compensation"

CONF_COMPENSATION = "compensation"
CONF_DATAPOINTS = "data_points"
CONF_DEGREE = "degree"
CONF_PRECISION = "precision"
CONF_POLYNOMIAL = "polynomial"

DATA_COMPENSATION = "compensation_data"

DEFAULT_DEGREE = 1
DEFAULT_NAME = "Compensation"
DEFAULT_PRECISION = 2
