"""Constants for device automations."""
CONF_IS_OFF = "is_off"
CONF_IS_ON = "is_on"
CONF_TOGGLE = "toggle"
CONF_TURN_OFF = "turn_off"
CONF_TURN_ON = "turn_on"
CONF_TURNED_OFF = "turned_off"
CONF_TURNED_ON = "turned_on"
