"""LiteJet constants."""

DOMAIN = "litejet"

CONF_EXCLUDE_NAMES = "exclude_names"
CONF_INCLUDE_SWITCHES = "include_switches"

PLATFORMS = ["light", "switch", "scene"]

CONF_DEFAULT_TRANSITION = "default_transition"
