"""Constants for the Modern Forms integration."""

DOMAIN = "modern_forms"

ATTR_IDENTIFIERS = "identifiers"
ATTR_MANUFACTURER = "manufacturer"

OPT_ON = "on"
OPT_SPEED = "speed"
OPT_BRIGHTNESS = "brightness"

# Services
SERVICE_SET_LIGHT_SLEEP_TIMER = "set_light_sleep_timer"
SERVICE_CLEAR_LIGHT_SLEEP_TIMER = "clear_light_sleep_timer"
SERVICE_SET_FAN_SLEEP_TIMER = "set_fan_sleep_timer"
SERVICE_CLEAR_FAN_SLEEP_TIMER = "clear_fan_sleep_timer"

ATTR_SLEEP_TIME = "sleep_time"
CLEAR_TIMER = 0
